# Copyright (c) str4d <str4d@mail.i2p>
# See COPYING for details.

from txi2p.plugins import I2PClientParser, I2PServerParser


i2pClientEndpointParser = I2PClientParser()
i2pServerEndpointParser = I2PServerParser()
